# -*- coding: utf-8 -*-

import os

import click

from . import SillyMolSpotter, MolSupplier


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

@click.group(context_settings=CONTEXT_SETTINGS)
def main():
    pass


@main.command(help='Create a spotter.', context_settings=CONTEXT_SETTINGS)
@click.option('-o', '--outfile', type=str, required=True,
              default=None, nargs=1, show_default=True, metavar='OUTFILE',
              help='Path to save the spotter to.')
@click.option('-r', '--radius', required=False, default=1, nargs=1,
              metavar='RADIUS', help='Radius of the Morgan fingerprinter to be used to identify silly bits.')
@click.option('-i', '--input', required=True, multiple=True, show_default=True, metavar='INFILE',
              help='Input file(s) to be parsed and added to the spotter.')
def create(outfile, radius, input):
    if isinstance(input, tuple):
        input = list(input)
    sms = SillyMolSpotter(radius)
    for file in input:
        sms.add_file(file)
    sms.save(outfile)


@main.command(help='Scores molecules with a spotter.', context_settings=CONTEXT_SETTINGS)
@click.option('-s', '--spotter', type=str, required=True,
              default=None, nargs=1, show_default=True, metavar='SPOTTER',
              help='Path to the saved the spotter or one of \'ChEMBL\', \'ExCAPE\' or \'Papyrus\'.')
@click.option('-i', '--input', required=True, multiple=True,
              metavar='INPUT', help='Input file(s) containing molecules to be scored.')
@click.option('-o', '--output', required=False, nargs=1,
              metavar='OUTPUT', help='Output file in which scores will be stored.')
def score(spotter, input, output):
    if output is not None and not os.access(os.path.dirname(output), os.W_OK):
        raise IOError(f'cannot write file to {output}')
    # try:
    sms = SillyMolSpotter.from_pretrained(spotter)
    # except:
    #     sms = SillyMolSpotter.from_file(spotter)
    if isinstance(input, tuple):
        input = list(input)
    if output is not None:
        output_handle = open(output, 'w')
    for file in input:
        with MolSupplier(file) as supplier:
            for mol in supplier:
                score = sms.score_mol(mol)
                if output is not None:
                    output_handle.write(str(score) + '\n')
                else:
                    print(score)
    if output is not None:
        output_handle.close()



# -*- coding: utf-8 -*-

from __future__ import annotations
import os
import json
import lzma
from typing import List, Union

from rdkit import Chem
from rdkit.rdBase import BlockLogs
from rdkit.Chem import AllChem, Draw
from PIL.ImageFile import ImageFile

from .io import MolSupplier


CHEMBL_SPOTTER = os.path.join(os.path.dirname(__file__), 'spotters', 'chembl_32.sp')
EXCAPE_SPOTTER = os.path.join(os.path.dirname(__file__), 'spotters', 'excapedb.sp')
PAPYRUS_SPOTTER = os.path.join(os.path.dirname(__file__), 'spotters', 'papyrus_05.6.sp')


class SillyMolSpotter:
    def __init__(self, fp_radius):
        """Initialize a spotter to identify silly molecules.

        :param fp_radius: radius of the Morgan fingerprinter
        """
        self.count_dict = {}
        self.radius = fp_radius

    @classmethod
    def from_file(cls, filepath: str) -> SillyMolSpotter:
        """Load a saved spotter."""
        if not os.path.isfile(filepath):
            raise FileNotFoundError(f'file does not exist {filepath}')
        with lzma.open(filepath) as infile:
            data = json.load(infile)
        sms = SillyMolSpotter(data['spotter_radius'])
        sms.count_dict = {int(x): int(y) for x, y in data['spotter_bits'].items()}
        return sms

    @classmethod
    def from_pretrained(cls, name) -> SillyMolSpotter:
        """Load a pretrained spotter.

        :param name: one of 'ChEMBL', 'ExCAPE' or 'Papyrus'
        """
        mapping = {'chembl': CHEMBL_SPOTTER,
                   'excape': EXCAPE_SPOTTER,
                   'papyrus': PAPYRUS_SPOTTER}
        if name.lower() not in mapping.keys():
            raise ValueError('Name of a pretrained spotter may only be one of \'ChEMBL\', \'ExCAPE\' or \'Papyrus\'')
        sms = SillyMolSpotter.from_file(mapping[name])
        return sms

    def add_mol(self, mol: Union[Chem.Mol, List[Chem.Mol]]) -> None:
        """Add sound molecule(s) to the reference set of the spotter.

        :param mol: a RDKit molecule or list of molecules
        """
        if not isinstance(mol, list):
            mol = [mol]
        block = BlockLogs()
        for mol_ in mol:
            if mol_ is not None:
                try:
                    fp = AllChem.GetMorganFingerprint(mol, self.radius)
                    for k, v in fp.GetNonzeroElements().items():
                        self.count_dict[k] = self.count_dict.get(k, 0) + v
                except:
                    pass
        del block

    def add_smiles(self, smiles: Union[str, List[str]]) -> None:
        """Add SMILES of molecule(s) to the reference set of the spotter.

        :param smiles: SMILES of a molecules or of a list of molecules.
        """
        if not isinstance(smiles, list):
            smiles = [smiles]
        block = BlockLogs()
        for notation in smiles:
            try:
                mol = Chem.MolFromSmiles(notation)
                if mol is not None:
                    fp = AllChem.GetMorganFingerprint(mol, self.radius)
                    for k, v in fp.GetNonzeroElements().items():
                        self.count_dict[k] = self.count_dict.get(k, 0) + v
            except:
                pass
        del block

    def add_file(self, filepath: str) -> None:
        """Add the molecular content of a file to the reference set of the spotter.

        :param filepath: path to a file containing molecular structures.
        """
        if not os.path.isfile(filepath):
            raise FileNotFoundError(f'file cannot be found: {filepath}')
        file_size = bool(os.path.getsize(filepath) // 2 ** 20)  # file is more than 1MB
        block = BlockLogs()
        with MolSupplier(filepath, show_progress=file_size) as supplier:
            for mol in supplier:
                if mol:
                    try:
                        fp = AllChem.GetMorganFingerprint(mol, self.radius)
                        for k, v in fp.GetNonzeroElements().items():
                            self.count_dict[k] = self.count_dict.get(k, 0) + v
                    except:
                        pass
        del block

    def score_mol(self, mol: Union[Chem.Mol, List[Chem.Mol]]) -> Union[float, List[float]]:
        """Score how silly the input molecule(s) is/are.

        :param mol: RDKit molecule or list of molecules
        """
        if not len(self.count_dict):
            raise ValueError('this spotter does not contain any bit, please add molecules before scoring')
        if not isinstance(mol, list):
            mol = [mol]
        scores = []
        for mol_ in mol:
            if mol_ is not None:
                fp = AllChem.GetMorganFingerprint(mol_, self.radius)
                on_bits = fp.GetNonzeroElements().keys()
                silly_bits = [x for x in [self.count_dict.get(x)
                                          for x in on_bits]
                              if x is None]
                score = len(silly_bits) / len(on_bits)
            else:
                score = 1
            scores.append(score)
        if len(scores) == 1:
            return scores[0]
        return scores

    def save(self, filepath: str) -> None:
        """Save a spotter to disk.

        :param filepath: path on disk to save the spotter to.
        """
        if not os.access(os.path.dirname(filepath), os.W_OK):
            raise IOError(f'cannot write file to {filepath}')
        with lzma.open(filepath, mode='wt', preset=9 | lzma.PRESET_EXTREME) as outfile:
            json.dump({'spotter_radius': self.radius,
                       'spotter_bits': self.count_dict},
                      outfile)

    def show_mol(self, mol: Chem.Mol, raise_error: bool = False, molsPerRow=3,
             subImgSize=(500, 300), useSVG=False, returnPNG=False, **kwargs) -> Union[str, ImageFile]:
        block = BlockLogs()
        if mol:
            # Extract bits information
            bi = {}
            _ = AllChem.GetMorganFingerprint(mol, radius=self.radius, bitInfo=bi, useCounts=False)
            # Collect atom and bond environments for drawing
            drawings = []
            for hash, atom_abb_envs in bi.items():
                skip = False
                atom_env = set()
                bond_env = []
                for atom, radius in atom_abb_envs:
                    if self.count_dict.get(hash) is not None:
                        skip = True
                        continue
                    bond_env.extend(Chem.FindAtomEnvironmentOfRadiusN(mol, self.radius, atom))
                    for bond in bond_env:
                        atom_env.add(mol.GetBondWithIdx(bond).GetBeginAtomIdx())
                        atom_env.add(mol.GetBondWithIdx(bond).GetEndAtomIdx())
                if not skip:
                    legend = f'{hash}'
                    drawings.append((tuple(atom_env), tuple(bond_env), legend))
            # Silly bits found
            if len(drawings):
                atom_highlights, bond_highlights, legends = zip(*drawings)
                return Draw.MolsToGridImage([mol] * len(drawings),
                                            molsPerRow=molsPerRow,
                                            subImgSize=subImgSize,
                                            legends=legends,
                                            highlightAtomLists=atom_highlights,
                                            highlightBondLists=bond_highlights,
                                            useSVG=useSVG,
                                            returnPNG=returnPNG,
                                            **kwargs)
            return 'No silly bit'
            # return Draw.MolToImage(mol, highlightAtoms=atoms_envs, highlightBonds=bonds_envs, legend='3026373406')
        elif raise_error:
            del block
            raise TypeError('smiles could not be parsed')
        del block



# -*- coding: utf-8 -*-

import bz2
import gzip
import io
import lzma
import re
from typing import Iterable, Optional, Tuple, Callable, Union

from rdkit import Chem
from rdkit.rdBase import BlockLogs
from rdkit.Chem import (ForwardSDMolSupplier, MaeMolSupplier, MolFromMol2Block, SmilesMolSupplierFromText,
                        SmilesMolSupplier)
from tqdm.auto import tqdm


class ForwardMol2MolSupplier:
    def __init__(self, fileobj: Union[str, io.TextIOBase],
                 sanitize: bool = True,
                 removeHs: bool = True,
                 cleanupSubstructures: bool = True):
        self.sanitize = sanitize
        self.removeHs = removeHs
        self.cleanupSubstructures = cleanupSubstructures
        self._buffer_size = 32768  # 32kB
        self._buffer = b''
        self._mol_delimiter = b'@<TRIPOS>MOLECULE'
        if isinstance(fileobj, str):
            self._open_supplier = True
            self.supplier = open(fileobj, 'rb')
        else:
            self._open_supplier = False
            self.supplier = fileobj

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _iterate(self):
        self._buffer = self.supplier.read(self._buffer_size)
        while True:
            i_seps = [x.start() for x in re.finditer(self._mol_delimiter, self._buffer) if x.start() != 0]
            if not i_seps:
                new_buffer = self.supplier.read(self._buffer_size)
                if len(new_buffer):
                    self._buffer += new_buffer
                else:
                    mol = MolFromMol2Block(self._buffer,
                                           self.sanitize,
                                           self.removeHs,
                                           self.cleanupSubstructures)
                    yield mol
                    break
            else:
                mol = MolFromMol2Block(self._buffer[:i_seps[0]])
                yield mol
                self._buffer = self._buffer[i_seps[0]:]
                del i_seps[0]

    def __iter__(self):
        if not hasattr(self, '_iterator'):
            self._iterator = self._iterate()
        for values in self._iterator:
            yield values

    def __next__(self):
        if not hasattr(self, '_iterator'):
            self._iterator = self._iterate()
        return next(self._iterator)

    def close(self):
        if self._open_supplier:
            self.supplier.close()


class ForwardSmilesMolSupplier:
    def __init__(self, fileobj: Union[str, io.TextIOBase],
                 delimiter: str = '\t',
                 smilesColumn: int = 0,
                 nameColumn: int = 1,
                 titleLine: bool = True,
                 sanitize: bool = True):
        self.delimiter = delimiter
        self.smilesColumn = smilesColumn
        self.nameColumn = nameColumn
        self.titleLine = titleLine
        self.sanitize = sanitize
        self._buffer_size = 32768  # 32kB
        self._buffer = b''
        self._mol_delimiter = '\n'
        if isinstance(fileobj, str):
            self._open_supplier = True
            self.supplier = None
            self._iterator = SmilesMolSupplier(fileobj, self.delimiter, self.smilesColumn, self.nameColumn,
                                               self.titleLine, self.sanitize)
        else:
            self._open_supplier = False
            self.supplier = fileobj

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _iterate(self):
        if self.titleLine:
            self.supplier.readline()
        self._buffer = self.supplier.read(self._buffer_size)
        while True:
            i_seps = [x.start() for x in re.finditer(self._mol_delimiter, self._buffer)]
            if not i_seps:
                new_buffer = self.supplier.read(self._buffer_size)
                if len(new_buffer):
                    self._buffer += new_buffer
                else:
                    if len(self._buffer):
                        block = BlockLogs() # Disable logger if no name column
                        mol = next(SmilesMolSupplierFromText(self._buffer, self._mol_delimiter, self.smilesColumn,
                                                             self.nameColumn, False, self.sanitize))
                        del block  # Disable logger if no name column
                        yield mol
                    break
            else:
                block = BlockLogs() # Disable logger if no name column
                mol = next(
                    SmilesMolSupplierFromText(self._buffer[:i_seps[0] + len(self._mol_delimiter)], self._mol_delimiter,
                                              self.smilesColumn, self.nameColumn, False, self.sanitize))
                del block  # Disable logger if no name column
                yield mol
                self._buffer = self._buffer[i_seps[0] + len(self._mol_delimiter):]
                del i_seps[0]

    def __iter__(self):
        if not hasattr(self, '_iterator'):
            self._iterator = self._iterate()
        for values in self._iterator:
            yield values

    def __next__(self):
        if not hasattr(self, '_iterator'):
            self._iterator = self._iterate()
        return next(self._iterator)

    def close(self):
        if self._open_supplier:
            del self._iterator
            self._iterator = None


class MolSupplier:
    # class properties
    valid_formats = ('smi', 'mae', 'sd', 'mol2', 'mol')
    valid_compression = ('lzma', 'zlib', 'bz2')

    def __init__(self, source: Union[str, io.TextIOBase, io.BufferedIOBase] = None,
                 supplier: Iterable[Chem.Mol] = None,
                 format: str = None,
                 compression: str = None, **kwargs):
        f"""Molecular supplier handling format and compression.
        :param source: filename or file-like object;
                       when using a context manager, file-like objects
                       are not closed upon exit
        :param supplier: molecular supplier (e.g. rdkit.Chem.ForwardSDMolSupplier)
        :param format: data format {self.valid_formats}
                       can be detected if source is a file name,
                       must be provided if source is a not file name,
                       ignored if supplier is not None
        :param compression: compression type {self.valid_compression}
                            can be detected if source is a file name,
                            ignored otherwise
        :param kwargs: keyworded arguments to be passed to the underlying supplier,
                       ignored if source is supplier
                       can also hold values for 'total' and 'show_progress'
                       to be considered when used as an iterable
        """
        # source is None
        if source is None and supplier is None:
            raise ValueError('source or supplier must be supplied')
        # Default attributes
        self._open_substream = False  # should a file be opened
        self.filename = None  # name of file to be opened
        self.open_fn = None  # function opening file and handling compression
        self._handle = None  # handle to opened file
        self._open_supplier = False  # should a supplier be opened
        self.supplier = None  # molecule supplier
        self.compression = None
        self.format = None
        self.kwargs = kwargs  # additional parameters for suppliers
        self._iter_total = self.kwargs.pop('total', None)
        self._iter_progress = self.kwargs.pop('show_progress', None)
        # Handle supplier
        if supplier is not None:
            self.supplier = supplier
        # source is a file name
        elif isinstance(source, str):
            self.filename = source
            self._open_substream = True
            self._open_supplier = True
            # Handle compressions
            if compression is not None:
                if compression not in self.valid_compression:
                    raise ValueError(f'compression must be one of {self.valid_compression}')
                self.compression = compression
            else:
                self.compression, self._trunc_filename = self._get_compression(self.filename)
            self.open_fn = self._get_compression_handler(self.compression)
            # Handle file types
            if format is not None:
                if format not in self.valid_formats:
                    raise ValueError(f'format must be one of {self.valid_formats}')
                self.format = format
            else:
                self.format = self._get_format(self._trunc_filename)
        # source is file-like object
        elif isinstance(source, (io.TextIOBase, io.BufferedIOBase)):
            if format is None:
                raise ValueError('format must be specified with text or binary readers')
            self._handle = source
            self._open_supplier = True
            self.format = format
        else:
            raise ValueError('source must either be filename or file-like object')
        # Create rdkit suppliers
        if self._open_substream:
            self._handle = self.open_fn(self.filename, 'rb')
        # if file name or file-like object
        if self._open_supplier:
            if self.format == 'smi':
                self.supplier = ForwardSmilesMolSupplier(self._handle, **self.kwargs)
            elif self.format == 'mae':
                self.supplier = MaeMolSupplier(self._handle, **self.kwargs)
            elif self.format in ['sd', 'mol']:
                self.supplier = ForwardSDMolSupplier(self._handle, **self.kwargs)
            elif self.format == 'mol2':
                self.supplier = ForwardMol2MolSupplier(self._handle, **self.kwargs)

    def set_start_progress_total(self, start: int=0, progress: bool=True, total: Optional[int]=None):
        """Set the start, progress and total for iterating through the supplier.
        :param start: starting value for generated identifiers while enumerating molecules
        :param progress: whether a progress bar should be displayed
        :param total: total number of molecules in the supplier
        """
        self._iter_start = start
        self._iter_total = total
        self._iter_progress = progress

    def _get_compression(self, filename: str) -> Tuple[Optional[str], str]:
        """Get compression type and stripped filename."""
        if filename.endswith('.xz'):
            return 'lzma', filename.rstrip('.xz')
        elif filename.endswith('.gz'):
            return 'zlib', filename.rstrip('.gz')
        elif filename.endswith('.bz2'):
            return 'bz2', filename.rstrip('.bz2')
        else:
            return None, filename

    def _get_compression_handler(self, compression_type) -> Callable:
        """Get function to deal with the compression."""
        if compression_type == 'lzma':
            return lzma.open
        elif compression_type == 'zlib':
            return gzip.open
        elif compression_type == 'bz2':
            return bz2.open
        elif compression_type is None:
            return open
        else:
            raise ValueError(f'type compression not handled: {compression_type}')

    def _get_format(self, filename) -> str:
        """Get file format from filename."""
        if filename.endswith('.smi'):
            return 'smi'
        elif filename.endswith('.mae'):
            return 'mae'
        elif filename.endswith(('.sd', '.sdf')):
            return 'sd'
        elif filename.endswith('.mol2'):
            return 'mol2'
        elif filename.endswith('.mol'):
            return 'mol'

    def _processed_mol_supplier(self) -> Iterable[Tuple[int, Chem.Mol]]:
        """Generator function that reads from a rdkit molecule supplier."""
        # handle showing progress
        if self._iter_progress:
            pbar = tqdm(self.supplier, total=self._iter_total, ncols=100)
        else:
            pbar = self.supplier
        for rdmol in pbar:
            if rdmol:
                yield rdmol
            else:
                continue

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __iter__(self):
        if not hasattr(self, '_iterator'):
            self._iterator = self._processed_mol_supplier()
        for values in self._iterator:
            yield values

    def __next__(self):
        if not hasattr(self, '_iterator'):
            self._iterator = self._processed_mol_supplier()
        return next(self._iterator)

    def close(self):
        if self._open_supplier:
            del self.supplier
            self.supplier = None
        if self._open_substream:
            self._handle.close()
